#include <string.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <stdlib.h>
#include <unistd.h>
#include <stdlib.h>
#include <openssl/rand.h>

#define SSH_AGENTC_LOCK		22
#define SSH_AGENTC_UNLOCK	23

#define SSH_AUTHSOCKET_ENV_NAME	"SSH_AUTH_SOCK"

#define PASS_BITS_ENV_NAME	"PASS_BITS"

#define LIBRARY_NAME		"libssh-agent-lock.so"

static char *pass = NULL;

static int locked = 0;

static int get_authentication_socket(void)
{
  const char *authsocket;
  int sock;
  struct sockaddr_un {
    short	sun_family;	/* AF_UNIX */
    char	sun_path[108];	/* path name (gag) */
  } sunaddr;

  authsocket = getenv(SSH_AUTHSOCKET_ENV_NAME);
  if (!authsocket)
    return -1;

  sunaddr.sun_family = AF_UNIX;
  strncpy(sunaddr.sun_path, authsocket, sizeof(sunaddr.sun_path));

  sock = socket(AF_UNIX, SOCK_STREAM, 0);
  if (sock < 0)
    return -1;

  if (connect(sock, (struct sockaddr *) &sunaddr, sizeof sunaddr) < 0) {
    close(sock);
    return -1;
  }

  return sock;
}

static void do_lock(int lock)
{
  int sock;
  int len;
  char buf[4];

  if (pass == NULL)
    return;

  sock = get_authentication_socket();

  if (sock < 0)
    return;

  len = 5 + strlen(pass);
  buf[0] = len >> 24;
  buf[1] = len >> 16;
  buf[2] = len >>  8;
  buf[3] = len;

  write(sock, buf, 4);

  buf[0] = lock ? SSH_AGENTC_LOCK : SSH_AGENTC_UNLOCK;

  write(sock, buf, 1);

  len = strlen(pass);
  buf[0] = len >> 24;
  buf[1] = len >> 16;
  buf[2] = len >>  8;
  buf[3] = len;

  write(sock, buf, 4);

  write(sock, pass, strlen(pass));

  close(sock);

  return;
}

static const char Base64[] =
	"ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/";

void _init(void)
{
  int bits = 256;
  int len, i;
  char *pass_bits = getenv(PASS_BITS_ENV_NAME);
  unsigned char *buf, *input, *output; 

  if (locked) {
    locked++;
    return;
  }

  locked++;

  if (getenv("LD_PRELOAD") && strstr(getenv("LD_PRELOAD"), LIBRARY_NAME))
    unsetenv("LD_PRELOAD");

  if (RAND_status() != 1)
    return;

  if (pass_bits)
    bits = atoi(pass_bits);

  len = (((bits+7)/8+2)/3)*3;
  buf = malloc(len);

  pass = malloc((len/3)*4+1);

  if (buf == NULL || pass == NULL || RAND_bytes(buf, len) <= 0) {
    free(buf);
    free(pass);
    pass = NULL;
    return;
  }

  input = buf;
  output = pass;

  for (i = len; i> 0; i-=3) {
    output[0] = Base64[input[0] >> 2];
    output[1] = Base64[((input[0] & 0x03) << 4) + (input[1] >> 4)];
    output[2] = Base64[((input[1] & 0x0f) << 2) + (input[2] >> 6)];
    output[3] = Base64[input[2] & 0x3f];

    input += 3;
    output += 4;
  }

  *output = '\0';

  for (i = 0; i < len; i++)
    buf[i] = 0;

  free(buf);

  do_lock(1);
}

void _fini(void)
{
  locked--;

  if (locked)
    return;

  do_lock(0);

  free(pass);
}
